import * as plugins from './plugins.js';
import { EventEmitter } from 'events';

export interface IProcessWrapperOptions {
  command: string;
  args?: string[];
  cwd: string;
  env?: NodeJS.ProcessEnv;
  name: string;
  logBuffer?: number; // Number of log lines to keep in memory (default: 100)
}

export interface IProcessLog {
  timestamp: Date;
  type: 'stdout' | 'stderr' | 'system';
  message: string;
}

export class ProcessWrapper extends EventEmitter {
  private process: plugins.childProcess.ChildProcess | null = null;
  private options: IProcessWrapperOptions;
  private logs: IProcessLog[] = [];
  private logBufferSize: number;
  private startTime: Date | null = null;
  
  constructor(options: IProcessWrapperOptions) {
    super();
    this.options = options;
    this.logBufferSize = options.logBuffer || 100;
  }
  
  /**
   * Start the wrapped process
   */
  public start(): void {
    this.addSystemLog('Starting process...');
    
    try {
      if (this.options.args && this.options.args.length > 0) {
        this.process = plugins.childProcess.spawn(this.options.command, this.options.args, {
          cwd: this.options.cwd,
          env: this.options.env || process.env,
          stdio: ['ignore', 'pipe', 'pipe'], // We need to pipe stdout and stderr
        });
      } else {
        // Use shell mode to allow a full command string
        this.process = plugins.childProcess.spawn(this.options.command, {
          cwd: this.options.cwd,
          env: this.options.env || process.env,
          stdio: ['ignore', 'pipe', 'pipe'], // We need to pipe stdout and stderr
          shell: true,
        });
      }
      
      this.startTime = new Date();
      
      // Handle process exit
      this.process.on('exit', (code, signal) => {
        this.addSystemLog(`Process exited with code ${code}, signal ${signal}`);
        this.emit('exit', code, signal);
      });
      
      // Handle errors
      this.process.on('error', (error) => {
        this.addSystemLog(`Process error: ${error.message}`);
        this.emit('error', error);
      });
      
      // Capture stdout
      if (this.process.stdout) {
        this.process.stdout.on('data', (data) => {
          const lines = data.toString().split('\n');
          for (const line of lines) {
            if (line.trim()) {
              this.addLog('stdout', line);
            }
          }
        });
      }
      
      // Capture stderr
      if (this.process.stderr) {
        this.process.stderr.on('data', (data) => {
          const lines = data.toString().split('\n');
          for (const line of lines) {
            if (line.trim()) {
              this.addLog('stderr', line);
            }
          }
        });
      }
      
      this.addSystemLog(`Process started with PID ${this.process.pid}`);
      this.emit('start', this.process.pid);
      
    } catch (error) {
      this.addSystemLog(`Failed to start process: ${error.message}`);
      this.emit('error', error);
      throw error;
    }
  }
  
  /**
   * Stop the wrapped process
   */
  public stop(): void {
    if (!this.process) {
      this.addSystemLog('No process running');
      return;
    }
    
    this.addSystemLog('Stopping process...');
    
    // First try SIGTERM for graceful shutdown
    if (this.process.pid) {
      try {
        process.kill(this.process.pid, 'SIGTERM');
        
        // Give it 5 seconds to shut down gracefully
        setTimeout(() => {
          if (this.process && this.process.pid) {
            this.addSystemLog('Process did not exit gracefully, force killing...');
            try {
              process.kill(this.process.pid, 'SIGKILL');
            } catch (error) {
              // Process might have exited between checks
            }
          }
        }, 5000);
      } catch (error) {
        this.addSystemLog(`Error stopping process: ${error.message}`);
      }
    }
  }
  
  /**
   * Get the process ID if running
   */
  public getPid(): number | null {
    return this.process?.pid || null;
  }
  
  /**
   * Get the current logs
   */
  public getLogs(limit: number = this.logBufferSize): IProcessLog[] {
    // Return the most recent logs up to the limit
    return this.logs.slice(-limit);
  }
  
  /**
   * Get uptime in milliseconds
   */
  public getUptime(): number {
    if (!this.startTime) return 0;
    return Date.now() - this.startTime.getTime();
  }
  
  /**
   * Check if the process is currently running
   */
  public isRunning(): boolean {
    return this.process !== null && typeof this.process.exitCode !== 'number';
  }
  
  /**
   * Add a log entry from stdout or stderr
   */
  private addLog(type: 'stdout' | 'stderr', message: string): void {
    const log: IProcessLog = {
      timestamp: new Date(),
      type,
      message,
    };
    
    this.logs.push(log);
    
    // Trim logs if they exceed buffer size
    if (this.logs.length > this.logBufferSize) {
      this.logs = this.logs.slice(-this.logBufferSize);
    }
    
    // Emit log event for potential handlers
    this.emit('log', log);
  }
  
  /**
   * Add a system log entry (not from the process itself)
   */
  private addSystemLog(message: string): void {
    const log: IProcessLog = {
      timestamp: new Date(),
      type: 'system',
      message,
    };
    
    this.logs.push(log);
    
    // Trim logs if they exceed buffer size
    if (this.logs.length > this.logBufferSize) {
      this.logs = this.logs.slice(-this.logBufferSize);
    }
    
    // Emit log event for potential handlers
    this.emit('log', log);
  }
}